/*****************************************************************************/
/*									     */
/*									     */
/*	Xpack version 1.00 -- module X-gfx.c				     */
/*									     */
/*	Drawing routines for the X window system.			     */
/*	Written by Michael Bischoff (mbi@mo.math.nat.tu-bs.de)		     */
/*	November-1994							     */
/*	see COPYRIGHT.xpack for Copyright details			     */
/*									     */
/*									     */
/*****************************************************************************/
#include "X-sok.h"

int DX = 8, DY = 8;

#ifndef DELAY
#define DELAY 50 /* time to sleep (in ms) between auto-moves & replay */
#endif

Display *dpy;
Window table = 0;
struct graphic graphic;
int gamegraphic = 1;

static GC gc;
static int resize_pending = 0;

#ifndef HAVE_USLEEP
#include <sys/time.h>
/* usleep emulation code taken from xboing-2.2 by Justin C. Kibell */

static int usleep(unsigned long usec) {
#ifdef SYSV
#ifdef __clipper__
    struct timeval tv;
    tv.tv_sec=((usec)/1000);
    tv.tv_usec=(((usec)%1000)*1000);
    select(1,NULL,NULL,NULL,&tv);
#else
    poll((struct poll *) 0, (size_t) 0, usec / 1000);   /* ms resolution */
#endif
#else
    struct timeval timeout;
    timeout.tv_usec = usec % (unsigned long) 1000000;
    timeout.tv_sec = usec / (unsigned long) 1000000;
    select(0, (void *) 0, (void *) 0, (void *) 0, &timeout);
#endif
    return 0;
}
#endif


void sync_and_wait(void) {
    XSync(dpy, 0);
    usleep(DELAY*1000);
}

void NewLevel(int levelnr) {
    SetTitle();
    if (table) {
	init_layout();
	AskWidgetForResize(graphic.width, graphic.height);
	cmd_Resize();
	if (gamegraphic)
	    refresh_screen();
	resize_pending = 1;
    }
}

void init_layout(void) {
    graphic.width = (P.t[0].width+2) * DX;
    graphic.height = (P.t[0].height+2) * DY;
    if (graphic.width < 144) {
	DX = DY = 1 + 244 / (P.t[0].width+2);
	graphic.width = (P.t[0].width+2) * DX;
	graphic.height = (P.t[0].height+2) * DY;
    } else if (graphic.height > 700) {
	DX = DY = 700 / (P.t[0].height+2);
	graphic.width = (P.t[0].width+2) * DX;
	graphic.height = (P.t[0].height+2) * DY;
    }
    /* printf("layout: DX = %d, width = %d\n", DX, P.t[0].width); */
}

static GC gcbg, gcfg, gcli, gcblack;
static int screen, fontheight;
static XFontStruct *message_font;

void cmd_Describe(void) {
    int i;
    for (i = 1; i <= P.totaltiles; ++i)
	if (P.position[i].x >= 0) {
	    struct tile *tp;
	    int x, y, w;
	    tp = P.t + P.position[i].tilenr;
	    w = XTextWidth(message_font, tp->name, strlen(tp->name));
	    if (w < DX * tp->width - 1 && fontheight < DY * tp->height - 1) {
		x = (P.position[i].x+1)*DX+1;
		y = (P.position[i].y+1)*DY+fontheight;
		y += (tp->height*DY-fontheight) / 2;
		x += (tp->width*DY-w) / 2;
		XDrawImageString(dpy, table, gcblack, x, y,
				 tp->name, strlen(tp->name));
	    }
	}
}

void init_gfx(unsigned long bgcolor, unsigned long fgcolor, unsigned long licolor) {
    XGCValues gcv;
    long gcflags;
    screen = DefaultScreen(dpy);
    gc = XDefaultGC(dpy, screen);
    if ((message_font = XLoadQueryFont(dpy, "fixed")) == NULL) {
	(void) fprintf(stderr,"can't get font fixed\n");
	exit(0);
    }

    gcv.background = WhitePixel(dpy, screen);
    gcv.graphics_exposures = True;
    gcflags = GCForeground | GCBackground | GCGraphicsExposures;
    gcv.foreground = bgcolor;
    gcbg = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
    gcv.foreground = fgcolor;
    gcfg = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
    gcv.foreground = licolor;
    gcli = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
    gcv.foreground = BlackPixel(dpy, screen);
    gcv.font = message_font->fid;
    gcflags |= GCFont;

    gcblack = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
    XSetLineAttributes(dpy, gcli,    1, LineSolid, CapButt, JoinMiter);
    XSetLineAttributes(dpy, gcblack, 1, LineSolid, CapButt, JoinMiter);
    fontheight = message_font->ascent + message_font->descent;
}

/* drag & drop code */

static Window moving_rect;
static int win_active = -1;
int offx, offy;		/* mouse coors */

static void draw_tile(void) {
    int x, y;
    Pixmap p;
    XSync(dpy, 0);
    p = XCreatePixmap(dpy, RootWindow(dpy, screen),
		      a.w*DX+1, a.h*DY+1, DefaultDepth(dpy, screen));
    XSync(dpy, 0);
    XFillRectangle(dpy, p, gcbg, 0, 0, a.w*DX, a.h*DY);
    XDrawRectangle(dpy, p, gcblack, 0, 0, a.w*DX, a.h*DY);
    for (y = 1; y < a.h; ++y)
	for (x = 0; x < a.w; ++x)
	    if (a.map[y-1][x] != a.map[y][x])
		XDrawLine(dpy, p, gcblack, x*DX, y*DY, (x+1)*DX, y*DY);
    for (x = 1; x < a.w; ++x)
	for (y = 0; y < a.h; ++y)
	    if (a.map[y][x-1] != a.map[y][x])
		XDrawLine(dpy, p, gcblack, x*DX, y*DY, x*DX, (y+1)*DY);
#if 0
    XCopyArea(dpy, p, moving_rect, DefaultGC(dpy, screen), 0, 0, a.w*DX+1, a.h*DY+1, 0, 0);
#else
    {
	long valuemask = CWBackPixmap;
	XSetWindowAttributes attributes;
	attributes.background_pixmap = p;
	XChangeWindowAttributes(dpy, moving_rect, valuemask, &attributes);
	XClearWindow(dpy, moving_rect);
    }
#endif
    XFreePixmap(dpy, p);
}

#ifdef SHAPE
#include <X11/extensions/shape.h>
static Pixmap shape_pixmap = 0;

static void add_shape(void) {
    char *pixels;
    int w, h;

    if (shape_pixmap)
	XFreePixmap(dpy, shape_pixmap);

    w = (a.w*DX+8) / 8;	/* add 1 for border */
    h = (a.h*DY+1);
    pixels = malloc_(w*h);
    memset(pixels, 0, w*h);
    /* pixels[0] = pixels[w] = pixels[2*w] = 0xff; */
    {   int x, y;
	for (y = 0; y < a.h; ++y)
	    for (x = 0; x < a.w; ++x)
		if (a.map[y][x]) {
		    /* set square (x,y) in bitmap */
		    
		    int xx, yy, maske;
		    for (xx = 0; xx <= DX; ++xx) {
			maske = 1 << ((x*DX + xx) & 7);
			for (yy = 0; yy <= DY; ++yy)
			    pixels[(y*DY+yy)*w + ((x*DX+xx) >> 3)] |= maske;
		    }
		}
    }
    /* memset(pixels, 0xff, w*h); */
    shape_pixmap = XCreateBitmapFromData(dpy, moving_rect, pixels,
					 a.w*DX+1, h);
    XShapeCombineMask(dpy, moving_rect, ShapeBounding, 0, 0, shape_pixmap,
		      ShapeSet);
    free_(pixels);
}
#else
static void add_shape(void) {
    ;		/* null funct; should draw shape */
}
#endif


void popdown_tile(void) {
    if (win_active == 1) {	
	win_active = 0;
	XUnmapWindow(dpy, moving_rect);
	/* printf("win inactivated\n"); */
    }
}

void resize_tile(void) {
    if (win_active == 1) {
	if (a.map) {
	    XResizeWindow(dpy, moving_rect, DX*a.w+1, DY*a.h+1); /* room for borders */
	    add_shape();
	    draw_tile();
	} else
	    XResizeWindow(dpy, moving_rect, DX*a.w, DY*a.h);
    }
}

void win_move(int x, int y) {
    if (win_active == 1)
	XMoveWindow(dpy, moving_rect, x-offx, y-offy);
}


static void make_win(void) {
    long valuemask = CWEventMask | CWBackPixel | CWBorderPixel;
    XSetWindowAttributes attributes;

    attributes.event_mask = 0; /* ButtonReleaseMask | ButtonMotionMask; */
    attributes.background_pixel = WhitePixel(dpy, screen);
    attributes.border_pixel = BlackPixel(dpy, screen);
    moving_rect = XCreateWindow(dpy, table, 20, 30, 10*DX+1, 10*DY+1,
				a.map ? 0 : 1, (int) CopyFromParent,
				InputOutput, (Visual *) CopyFromParent,
				valuemask, &attributes);
    win_active = 0;	/* initialized, but unmapped */
}

void popup_active(int basex, int basey, int x, int y) {
    if (win_active < 0)
	make_win();
    offx = x - basex;
    offy = y - basey;
    win_active = 1;
    if (a.map) {
	XMoveResizeWindow(dpy, moving_rect, DX*basex, DY*basey,
			  a.w*DX+1, a.h*DY+1);
	
	add_shape();
	XMapWindow(dpy, moving_rect);
	draw_tile();
    } else {
	XMoveResizeWindow(dpy, moving_rect, DX*basex, DY*basey,
		      a.w*DX-1, a.h*DY-1);
	XMapWindow(dpy, moving_rect);
    }
}

#if 0
    if (graphic.is_color && pixel != BlackPixel(dpy, screen)) {
	gcv.foreground = pixel;
	dashedlinegc = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
	XSetLineAttributes(dpy, dashedlinegc, LINEWIDTH, LineSolid, CapButt, JoinMiter);
    } else {
	gcv.foreground = BlackPixel(dpy, screen);
	dashedlinegc = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
	XSetLineAttributes(dpy, dashedlinegc, LINEWIDTH, LineOnOffDash, CapButt, JoinMiter);
    }
    gcv.foreground = WhitePixel(dpy, screen);
    gcv.background = BlackPixel(dpy, screen);
    gcv.graphics_exposures = True;
    gcflags = GCForeground | GCBackground | GCGraphicsExposures;
    whitelinegc = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);

    XSetLineAttributes(dpy, whitelinegc, LINEWIDTH, LineSolid, CapButt, JoinMiter);
    stipplemap = XCreateBitmapFromData(dpy, RootWindow(dpy, screen),
				       stipple_bits, stipple_width, stipple_height);
    gcflags = GCForeground | /* GCBackground | */ GCGraphicsExposures |
	GCFillStyle | GCStipple;
    gcv.foreground = BlackPixel(dpy, screen);
    gcv.background = WhitePixel(dpy, screen);
    gcv.stipple = stipplemap;
    gcv.fill_style = FillStippled;
    stipplegc = XCreateGC(dpy, RootWindow(dpy, screen), gcflags, &gcv);
#endif



static void line(int a, int b, int x, int y, int dx, int dy) {
    x *= DX; y*= DY; dx *= DX; dy *= DY;
    if (a != b)
	XDrawLine(dpy, table, gcblack, x, y, x+dx, y+dy);
    else
	if (a == 0) {
	    if (dx)
		++x, dx -= 2;
	    else
		++y, dy -= 2;
	    XDrawLine(dpy, table, gcbg, x, y, x+dx, y+dy);
	}
}

static void do_redraw(int x, int y) {
    GC gc;
    if (x > P.t[0].width+1 || y > P.t[0].height+1)
	return;
    /* printf("painting %d at %d,%d\n", P.board[y][x], x, y); */
    switch (P.board[y][x]) {
    case -1:
	gc = gcfg;
	break;
    case 0:
	gc = gcli;
	break;
    default:
	gc = gcbg;
    }
    XFillRectangle(dpy, table, gc, x*DX, y*DY, DX+1, DY+1);
    if (x > 0)
	line(P.board[y][x], P.board[y][x-1], x, y, 0, 1);
    if (x <= P.t[0].width)
	line(P.board[y][x], P.board[y][x+1], x+1, y, 0, 1);
    if (y > 0)
	line(P.board[y][x], P.board[y-1][x], x, y, 1, 0);
    if (y <= P.t[0].height)
	line(P.board[y][x], P.board[y+1][x], x, y+1, 1, 0);
}

static void dotPaint(int minx, int miny, int maxx, int maxy) {
    int x, y;
    if (minx > maxx || miny > maxy) {
	int h;
	h = minx; minx = maxx; maxx = h;
	h = miny; miny = maxy; maxy = h;
    }
    x = minx-1;
    do {
	++x;
	for (y = miny; y <= maxy; ++y) {
	    do_redraw(x, y);
	}
    } while (x != maxx);
}

/* doPaint: just request it */
void doPaint(int minx, int miny, int maxx, int maxy) {
#if 0
    XClearArea(dpy, table, minx*DX, miny*DY, (maxx+1)*DX-1, (maxy+1)*DY-1, True);
#else
    dotPaint(minx, miny, maxx, maxy);
#endif
}

void redraw_table(XExposeEvent *xev) {
    int minx, miny, maxx, maxy;
    if (resize_pending) {
	if (xev->count)
	    return;
	/* after a resize, do a complete redraw */
	XClearArea(dpy, table, 0, 0, 0, 0, False);
	dotPaint(1, 1, game.numcols-2, game.numrows-2);
	resize_pending = 0;
    } else {
	minx = xev->x / DX;
	miny = xev->y / DY;
	maxx = (xev->x + xev->width - 1) / DX;
	maxy = (xev->y + xev->height - 1) / DY;
	dotPaint(minx, miny, maxx, maxy);
    }
}
